
#include once "definitions.bi"
#include once "types.bi"
#include once "utils.bi"
#include once "board.bi"

type TArtificialChessPlayer
  as byte minDepth = 3
  as byte maxDepth = 5
  as TBoard playBoard, startBoard
  as TMoveList list
  declare sub Initialize(byval as string = "")
  declare function RecursiveEvaluation(byref as TBoard, byval as byte, byval as byte, byval as short) as short
  declare function UserMove(byval as string, byval as byte = cQueenValue) as byte
  declare function InitialState(byref as TBoard, byval as byte) as byte
  declare function SearchBestMove(byval as short) as byte
  declare function BestMove(byref as byte) as string
end type

sub TArtificialChessPlayer.Initialize(byval s as string = "")
  startBoard.Initialize("")
  playBoard.Initialize(s)
end sub

function TArtificialChessPlayer.RecursiveEvaluation( _
  byref aBoard as TBoard, _
  byval aColor as byte, _
  byval aDepth as byte, _
  byval aAlpha as short _
) as short

  dim as byte i, halt
  dim as short beta, value
  dim as TBoard brd

  aBoard.GenerateAllMoves()
  
  beta = -32000 * aBoard.activeColor * aColor
  
  i = 0
  halt = false

  while (i < aBoard.list.count) and not halt
    i += 1
    brd = aBoard
    
    with brd
      .MovePiece(.list.array(i).fromSquare, .list.array(i).toSquare)
    end with

    if ( _
      (aDepth >= minDepth) and _
      (aBoard.board(aBoard.list.array(i).toSquare) = cNothingValue) _
   ) _
    or (aDepth = maxDepth) then
      value = brd.PositionValue(aColor)
    else
      value = RecursiveEvaluation(brd, aColor, aDepth + 1, beta)
    end if
    
    if aBoard.activeColor = aColor then
      if value > beta then beta = value
      if beta > aAlpha then halt = true
    else
      if value < beta then beta = value
      if beta < aAlpha then halt = true
    end if

    aBoard.list.array(i).value = value
  wend
  
  return beta
end function

#define in_ah(c) c >= asc("a")) and (c <= asc("h")
#define in_18(c) c >= asc("1")) and (c <= asc("8")

function TArtificialChessPlayer.UserMove( _
  byval arg as string, _
  byval aPromotion as byte = cQueenValue _
) as byte

  dim as byte fromSquare, toSquare
  dim as byte result
  dim as integer move
  dim brd as TBoard
  dim aux as string = arg
  dim p as string
  
  if (len(arg) >= 4) andalso _
  (in_ah(arg[0])) and (in_18(arg[1])) and _
  (in_ah(arg[2])) and (in_18(arg[3])) then
    aux[0] -= asc("a") - asc("1")
    aux[2] -= asc("a") - asc("1")
  else
    return false
  end if
  
  result = false
  if len(aux) = 5 then
    p = mid(aux, 5, 1)
    if (p <> "q") and (p <> "Q") then
      ToLog("** The program only knows the promotion in queen!")
    end if
    aux = mid(aux, 1, 4)
  end if
  move = valint(aux)
  playBoard.GenerateAllMoves
  for i as integer = 1 to playBoard.list.count
    if move = playBoard.list.array(i).fromSquare * 100 + playBoard.list.array(i).toSquare then
      result = true
      exit for
    end if
  next i
  
  if not result then return result
  
  fromSquare = move \ 100
  toSquare = move mod 100
  brd = playBoard
  brd.MovePiece(fromSquare, toSquare)
  brd.activeColor = -1 * brd.activeColor
  
  if brd.Check then
    result = false
  else
    playBoard.MovePiece(fromSquare, toSquare, aPromotion)
    playBoard.halfmoveCount += 1
  end if
  
  return result
end function

function TArtificialChessPlayer.InitialState(byref aBoard as TBoard, byval aSquare as byte) as byte
  if aBoard.board(aSquare) = startBoard.board(aSquare) then
    return true
  else
    return false
  end if
end function

function TArtificialChessPlayer.SearchBestMove(byval aRecursiveEvaluation as short) as byte
  dim brd as TBoard
  dim as byte i, j
  dim as byte piece
  dim as integer max
  dim as byte index
  
  list.count = 0
  for i = 1 to playBoard.list.count
    if playBoard.list.array(i).value = aRecursiveEvaluation then
      list.count += 1
      list.array(list.count).fromSquare = playBoard.list.array(i).fromSquare
      list.array(list.count).toSquare = playBoard.list.array(i).toSquare
      list.array(list.count).value = 0
    end if
  next i
  
  max = 0
  index = 1
  for i = 1 to list.count
    brd = playBoard
    
    with list.array(i)
      .value = 0
      piece = abs(brd.board(.fromSquare))
      
      if InitialState(brd, .fromSquare) then
        .value += 5
        if piece = cPawnValue then .value += 2
      end if
      
      if (piece = cKingValue) then .value -= 10
      
      if (brd.halfmoveCount < 32) then
        if (piece = cPawnValue) _
        or (piece = cBishopValue) _
        or (piece = cKnightValue) then .value += 20
      end if
      
      if (.fromSquare \ 10 = 1) _
      or (.fromSquare \ 10 = 8) _
      or (.fromSquare mod 10 = 1) _
      or (.fromSquare mod 10 = 8) _
      then .value += 2
      
      if (.toSquare \ 10 = 1) _
      or (.toSquare \ 10 = 8) _
      or (.toSquare mod 10 = 1) _
      or (.toSquare mod 10 = 8) _
      then .value -= 2
      
      brd = playBoard
      brd.MovePiece(.fromSquare, .toSquare)
      if InitialState(brd, .toSquare) then .value -= 10
      brd.activeColor = -1 * brd.activeColor
      brd.GenerateMoves()
      
      for j = 1 to brd.list.count
        .value += 1
        if brd.board(brd.list.array(j).toSquare) <> cNothingValue then .value += 1
      next j
      
      brd.activeColor = -1 * brd.activeColor
      brd.GenerateMoves
      
      for j = 1 to brd.list.count
        .value -= 1
        if brd.board(brd.list.array(j).toSquare) <> cNothingValue then .value -= 1
      next j
      
      if .value >= max then
        max = .value
        index = i
      end if
    end with
  next i
  
  return index
end function

function TArtificialChessPlayer.BestMove(byref aFlag as byte) as string
  dim as byte i, alreadyInCheck
  dim as string result 
  dim promotion as boolean
  
  i = SearchBestMove(RecursiveEvaluation(playBoard, playBoard.activeColor, 1, 32000))
  
  playBoard.halfmoveCount += 1
  alreadyInCheck = playBoard.Check
  
  with list.array(i)
    promotion = (abs(playBoard.board(.fromSquare)) = cPawnValue) andalso ((.toSquare mod 10 = 1) or (.toSquare mod 10 = 8))
  end with
  
  playBoard.MovePiece(list.array(i).fromSquare, list.array(i).toSquare)
  result = MoveToStr(list.array(i).fromSquare, list.array(i).toSquare)
  
  if promotion then result += "q"
  
  playBoard.activeColor = -1 * playBoard.activeColor
  if playBoard.Check then
    if alreadyInCheck then
      aFlag = 3 ' Illegal move (checkmate)
    else
      aFlag = 2 ' Illegal move (stalemate)
    end if
  else
    playBoard.activeColor = -1 * playBoard.activeColor
    if playBoard.Check then
      aFlag = 1 ' Check
    else
      aFlag = 0
    end if
  end if
  
  return result
end function
